unit BugCommon;

interface

uses
  Classes, // for TPoint, TAlignment
  Windows,
  Graphics; // Graphics should ALWAYS come after Windows, to get the right (Graphics) definition of TBitmap

function AnotherColour: TColor;
function ProportionFromToColour(const Factor: single; const FromColour, ToColour: TColor): TColor;

procedure ClearRectangle(ACanvas: TCanvas; const ARect: TRect; const AColour: TColor);
procedure GraduClearRectangle(ACanvas: TCanvas; const ARect: TRect; const FromColour, ToColour: TColor; const ClearLeftToRight: boolean = false);
procedure PyramidClearRectangle(ACanvas: TCanvas; const ARect: TRect; const OutsideColour, CentreColour: TColor);

procedure LineWithEnds(ACanvas: TCanvas; const FromPoint, ToPoint: TPoint; const WantDot, WantArrow: boolean; const ArrowIsSolid: boolean = false);

procedure SomeLines( ACanvas: TCanvas );

function MultiLineTextExtent(ACanvas: TCanvas; const Text: string): TSize;
procedure MultiLineTextOut(ACanvas: TCanvas; const X, Y: integer; const Text: string; const Alignment: TAlignment);

procedure FlipBitmap(ABitmap: TBitmap);
procedure MirrorBitmap(ABitmap: TBitmap);

procedure BugBorder(ACanvas: TCanvas; const ARect: TRect);

procedure Circle(ACanvas: TCanvas; const Centre: TPoint; const Radius: integer); overload;
procedure Circle(ACanvas: TCanvas; const Centre: TPoint; const ACircumfPoint: TPoint); overload;
procedure Circle(ACanvas: TCanvas; const A, B, C: TPoint); overload;

function DistanceBetweenPoints(const A, B: TPoint): integer; // should really be a real for accuracy, but an integer will do for here
function BearingFromPointToPoint(const A, B: TPoint): integer; // result is 0..359 - should really be a real for accuracy, but an integer will do for here

implementation

uses
  Math; // for Arctan2

{$WRITEABLECONST ON}  // this unit uses Typed Constants as Variables, so we must be able to write to them


// this Round function, local to this unit, will hide the Delphi function and so give a proper rounding!
function Round(const ANumber: extended): integer;
{the Delphi Round function rounds a value of .5 up OR down as it feels!}
{this version *definitely* rounds a .5 value to the next higher value}
begin
  Result := Trunc(Abs(ANumber) + 0.5);
  if ANumber < 0
  then Result := - Result;
end; // function Round


function AnotherColour: TColor;
const
  Colours: array[0..9] of TColor // dark and bright, R, G, B, Y + W, K
          = (clMaroon, clRed, clGreen, clLime, clNavy, clBlue, clOlive, clYellow, clWhite, clBlack);
  Index: integer = -1;
begin
  Index := Succ(Index) mod Length(Colours);
  Result := Colours[Index];
end; //function AnotherColour


function ProportionFromToColour(const Factor: single; const FromColour, ToColour: TColor): TColor;
// Factor should ideally be in range 0.0 .. 1.0, but the code can cope with wider range
// Will cause problems if System colours (ie code numbers) are passed in rather than real (RGB) colours
// - it would be better to use the ColorToRGB function on the passed-in colours before getting their RGB values
  function TruncateInto0to255(const ANumber: real): byte; // an in-line function
  begin
    if ANumber > 255 then Result := 255
    else if ANumber < 0 then Result := 0
    else Result := Round(ANumber);
  end; // of in-line function
var
  R_From, R_To, R_Result,
  G_From, G_To, G_Result,
  B_From, B_To, B_Result: byte;
  R_Change,
  G_Change,
  B_Change: integer;
begin
  R_From := GetRValue(FromColour); // FromColour and $000000FF;
  G_From := GetGValue(FromColour); // (FromColour and $0000FF00) shr 8;
  B_From := GetBValue(FromColour); // (FromColour and $00FF0000) shr 16;
  R_To   := GetRValue(ToColour);   // ToColour and $000000FF;
  G_To   := GetGValue(ToColour);   // (ToColour and $0000FF00) shr 8;
  B_To   := GetBValue(ToColour);   // (ToColour and $00FF0000) shr 16;
  R_Change := R_To - R_From;
  G_Change := G_To - G_From;
  B_Change := B_To - B_From;
  R_Result := TruncateInto0to255(R_From + (Factor * R_Change));
  G_Result := TruncateInto0to255(G_From + (Factor * G_Change));
  B_Result := TruncateInto0to255(B_From + (Factor * B_Change));
  Result := RGB( R_Result, G_Result, B_Result );
end; {function ProportionFromToColour}

procedure ClearRectangle(ACanvas: TCanvas; const ARect: TRect; const AColour: TColor);
var
  BrushWas: TBrush;
begin
  BrushWas:= TBrush.Create;
  try {..finally}
    with ACanvas
    do begin
         BrushWas.Assign(Brush);
         Brush.Style := bsSolid;
         Brush.Color := AColour;
         FillRect(ARect);
         Brush.Assign(BrushWas);
       end; //with ACanvas
  finally
    BrushWas.Free;
  end; {try..finally}
end; //procedure ClearRectangle

procedure GraduClearRectangle(ACanvas: TCanvas; const ARect: TRect; const FromColour, ToColour: TColor; const ClearLeftToRight: boolean = false);
var
  Y, RectHeight, RectWidth: integer;
  PenWas: TPen;
begin
  PenWas := TPen.Create;
  try
    PenWas.Assign(ACanvas.Pen);
    ACanvas.Pen.Width := 1;
    ACanvas.Pen.Style := psSolid;
    RectHeight := ARect.Bottom - ARect.Top;
    RectWidth := ARect.Right - ARect.Left;
    if ClearLeftToRight
    then for Y:= 0 to Pred(RectWidth)
         do begin
               ACanvas.Pen.Color := ProportionFromToColour(Y / RectWidth, FromColour, ToColour);
               ACanvas.PolyLine([Point(ARect.Left + Y, ARect.Top), Point(ARect.Left + Y, ARect.Bottom)]);
            end
    else // clear Top to Bottom
          for Y:= 0 to Pred(RectHeight)
          do begin
               ACanvas.Pen.Color := ProportionFromToColour(Y / RectHeight, FromColour, ToColour);
               ACanvas.PolyLine([Point(ARect.Left, ARect.Top + Y), Point(ARect.Right, ARect.Top + Y)]);
             end;
    ACanvas.Pen.Assign(PenWas);
  finally
    PenWas.Free;
  end;
end; // procedure GraduClearRectangle

procedure LineWithEnds(ACanvas: TCanvas; const FromPoint, ToPoint: TPoint; const WantDot, WantArrow: boolean; const ArrowIsSolid: boolean = false);
var
  LineDirection, Direction1, Direction2,  //radians
  SinDirection1, CosDirection1,
  SinDirection2, CosDirection2: extended;
  PenWas: TPen;
  BrushWas: TBrush;
  Arrow1, Arrow2: TPoint;
const
  ARROW_LENGTH = 10;    //pixels
  ARROW_ANGLE = PI / 6; //radians for 30 degrees
  CIRCLE_RADIUS = 4;    //pixels
begin
  ACanvas.PolyLine([FromPoint, ToPoint]);
  if WantDot // on from-end
  then begin
          PenWas:= TPen.Create;
          BrushWas:= TBrush.Create;
          try
            PenWas.Assign(ACanvas.Pen);
            BrushWas.Assign(ACanvas.Brush);
            ACanvas.Pen.Style := psSolid;
            ACanvas.Brush.Style := bsSolid;
            ACanvas.Brush.Color := ACanvas.Pen.Color;
            ACanvas.Ellipse(FromPoint.X - CIRCLE_RADIUS, FromPoint.Y - CIRCLE_RADIUS, FromPoint.X + CIRCLE_RADIUS, FromPoint.Y + CIRCLE_RADIUS);
            ACanvas.Pen.Assign(PenWas);
            ACanvas.Brush.Assign(BrushWas);
          finally
            PenWas.free;
            BrushWas.free;
          end;
       end;
  if WantArrow // on to-end
  then begin
         LineDirection := ArcTan2(ToPoint.X - FromPoint.X, FromPoint.Y - ToPoint.Y);
         Direction1 := LineDirection - PI - ARROW_ANGLE;
         Direction2 := LineDirection - PI + ARROW_ANGLE;
         SinCos(Direction1, SinDirection1, CosDirection1);
         SinCos(Direction2, SinDirection2, CosDirection2);
         Arrow1 := Point(ToPoint.X + Round(ARROW_LENGTH * SinDirection1),
                         ToPoint.Y - Round(ARROW_LENGTH * CosDirection1) );
         Arrow2 := Point(ToPoint.X + Round(ARROW_LENGTH * SinDirection2),
                         ToPoint.Y - Round(ARROW_LENGTH * CosDirection2) );
         if ArrowIsSolid
         then begin
                PenWas:= TPen.Create;
                BrushWas:= TBrush.Create;
                try
                  PenWas.Assign(ACanvas.Pen);
                  BrushWas.Assign(ACanvas.Brush);
                  ACanvas.Pen.Style := psSolid;
                  ACanvas.Brush.Style := bsSolid;
                  ACanvas.Brush.Color := ACanvas.Pen.Color;
                  ACanvas.Polygon([Arrow1, ToPoint, Arrow2]); // draws the arrowhead as a solid shape
                  ACanvas.Pen.Assign(PenWas);
                  ACanvas.Brush.Assign(BrushWas);
                finally
                  PenWas.free;
                  BrushWas.free;
                end;
              end
         else ACanvas.PolyLine([Arrow1, ToPoint, Arrow2]); // draws the arrowhead in same penstyle
       end;
end; //procedure LineWithEnds

procedure SomeLines( ACanvas: TCanvas );
var
  X, Y,
  R, G, B,
  index: integer;
begin
  ACanvas.Pen.Width := 2;
  X:= 50; Y:= 50;
  R:= 10; G:= 10; B:= 10;
  for index := 0 to 99
  do begin
       inc(X, 2); inc(Y, 2);
       inc(R, 2); inc(G, 2); inc(B, 2);
       ACanvas.Pen.Color := RGB(R, G, B);
       ACanvas.Polyline([Point(X, Y), Point(X + 100, Y)]);
     end;
end; //procedure SomeLines

procedure PyramidClearRectangle(ACanvas: TCanvas; const ARect: TRect; const OutsideColour, CentreColour: TColor);
var
  X, Y, RectHeight, RectWidth, RWd2, RHd2, Xoffset: integer;
  PenWas: TPen;
begin
  PenWas := TPen.Create;
  try
    PenWas.Assign(ACanvas.Pen);
    ACanvas.Pen.Width := 1;
    ACanvas.Pen.Style := psSolid;
    RectHeight := ARect.Bottom - ARect.Top;
    RectWidth := ARect.Right - ARect.Left;
    RWd2 := RectWidth div 2;
    RHd2 := RectHeight div 2;
    for X:= 0 to RWd2
    do begin
         ACanvas.Pen.Color := ProportionFromToColour(X / RWd2, OutsideColour, CentreColour);
         ACanvas.PolyLine([Point(ARect.Left + X, ARect.Top), Point(ARect.Left + X, ARect.Bottom)]);
         ACanvas.PolyLine([Point(ARect.Right - X, ARect.Top), Point(ARect.Right - X, ARect.Bottom)]);
       end;
    if RWd2 * 2 <> RectWidth
    then ACanvas.PolyLine([Point(Succ(RWd2), ARect.Top), Point(Succ(RWd2), ARect.Bottom)]);
    for Y:= 0 to RHd2
    do begin
         Xoffset := Round(Y * RWd2 / RHd2);
         ACanvas.Pen.Color := ProportionFromToColour(Y / RHd2, OutsideColour, CentreColour);
         ACanvas.PolyLine([Point(ARect.Left + Xoffset, ARect.Top + Y), Point(ARect.Right - Xoffset, ARect.Top + Y)]);
         ACanvas.PolyLine([Point(ARect.Left + Xoffset, ARect.Bottom - Y), Point(ARect.Right - Xoffset, ARect.Bottom - Y)]);
       end;
    ACanvas.Pen.Assign(PenWas);
  finally
    PenWas.Free;
  end;
end; //procedure PyramidClearRectangle

procedure MultiLineTextOut(ACanvas: TCanvas; const X, Y: integer; const Text: string; const Alignment: TAlignment);
var
  TheSize: TSize;
  TextDrawingRect: TRect;
  Flags: UINT;
begin
  TheSize:= MultiLineTextExtent(ACanvas, Text);
  Flags:= DT_NOCLIP; // speeds the drawing by not worrying about clipping to the rectangle
  case Alignment of
    taLeftJustify:  Flags:= DT_NOCLIP;
    taCenter:       Flags:= DT_NOCLIP or DT_CENTER;
    taRightJustify: Flags:= DT_NOCLIP or DT_RIGHT;
  end; //case
  TextDrawingRect := Rect(X, Y, X + TheSize.cX, Y + TheSize.cY);
  DrawText(ACanvas.handle, pchar(Text), -1, TextDrawingRect, Flags); // Windows API call, so it needs the Canvas handle
end; //procedure MultiLineTextOut

function MultiLineTextExtent(ACanvas: TCanvas; const Text: string): TSize;
var
  TextBoundingRect: TRect;
begin
  TextBoundingRect := Rect(0,0,0,0);
  DrawText(ACanvas.handle, pchar(Text), -1, TextBoundingRect, DT_CALCRECT); // Windows API call, so it needs the Canvas handle
                                                                            // doesnt draw the text, just calculates how much space it's gonna take
  Result.cx := TextBoundingRect.Right;
  Result.cy := TextBoundingRect.Bottom;
end; //procedure MultiLineTextOut

procedure MirrorBitmap(ABitmap: TBitmap);
// mirror is swap Left/Right sides
type
  TRGBQuadArray= packed array[WORD] of TRGBQuad; // redefine the array as WORD rather than BYTE
  PRGBQuadArray= ^TRGBQuadArray;
var
  bmpIn, bmpOut: TBitmap;
  pScanLineIn, pScanLineOut: PRGBQuadArray;
  X, Y: integer;
  TransparentWas: boolean;
  TransparentColourWas: TColor;
begin
  TransparentWas:= ABitmap.Transparent;
  TransparentColourWas:= ABitmap.TransparentColor;
  bmpIn:= TBitmap.Create;
  bmpOut:= TBitmap.Create;
  try
    bmpIn.Assign(ABitmap);
    bmpIn.PixelFormat := pf32bit;
    bmpOut.PixelFormat := pf32bit;
    bmpOut.Height := bmpIn.Height;
    bmpOut.Width := bmpIn.Width;
    for Y:= 0 to Pred(bmpIn.Height)
    do begin
         pScanLineIn := bmpIn.ScanLine[Y];
         pScanLineOut:= bmpOut.ScanLine[Y];
         for X:= 0 to Pred(bmpIn.Width)
         do begin
              pScanLineOut^[Pred(bmpIn.Width - X)] := pScanLineIn^[X];
            end; // for each pixel-quad in this Scanline
       end; // for each Scanline
    ABitmap.Assign(bmpOut);
    ABitmap.TransparentColor:= TransparentColourWas;
    ABitmap.Transparent := TransparentWas;
  finally
    bmpIn.free;
    bmpOut.free;
  end;
end;

procedure FlipBitmap(ABitmap: TBitmap);
// flip is swap Top/Bottom
var
  bmpIn, bmpOut: TBitmap;
  Y, NumBytes: integer;
  TransparentWas: boolean;
  TransparentColourWas: TColor;
begin
  TransparentWas:= ABitmap.Transparent;
  TransparentColourWas:= ABitmap.TransparentColor;
  bmpIn:= TBitmap.Create;
  bmpOut:= TBitmap.Create;
  try
    bmpIn.Assign(ABitmap);
    bmpIn.PixelFormat := pf32bit;
    bmpOut.PixelFormat := pf32bit;
    bmpOut.Height := bmpIn.Height;
    bmpOut.Width := bmpIn.Width;
    NumBytes := bmpIn.Width * 4;
    for Y:= 0 to Pred(bmpIn.Height)
    do CopyMemory(bmpOut.Scanline[Pred(bmpIn.Height) - Y], bmpIn.Scanline[Y], NumBytes);
    ABitmap.Assign(bmpOut);
    ABitmap.TransparentColor:= TransparentColourWas;
    ABitmap.Transparent := TransparentWas;
  finally
    bmpIn.free;
    bmpOut.free;
  end;
end;

procedure BugBorder(ACanvas: TCanvas; const ARect: TRect);
// draws a rectangle with quarter-arcs as the corners
// the size of the arcs is fixed in this example, but could be a parameter passed in.
var
  // the points at the ends of the 4 lines of the rectangular part of the border
  TL, TR, // Top line, Left and Right ends
  RT, RB, // Right line, Top and Bottom ends
  BR, BL, // Bottom line, Right and Left ends
  LB, LT  // Left line, Bottom and Top ends
        : TPoint;
const
  ARC_RADIUS = 15;
begin
  if ((ARect.Right - ARect.Left) > (2 * ARC_RADIUS))
     and ((ARect.Bottom - ARect.Top) > (2 * ARC_RADIUS))
  then begin
        TL := Point(ARect.Left + ARC_RADIUS, ARect.Top);
        TR := Point(ARect.Right - ARC_RADIUS, ARect.Top);
        RT := Point(ARect.Right, ARect.Top + ARC_RADIUS);
        RB := Point(ARect.Right, ARect.Bottom - ARC_RADIUS);
        BR := Point(ARect.Right - ARC_RADIUS, ARect.Bottom);
        BL := Point(ARect.Left + ARC_RADIUS, ARect.Bottom);
        LB := Point(ARect.Left, ARect.Bottom - ARC_RADIUS);
        LT := Point(ARect.Left, ARect.Top + ARC_RADIUS);
        with ACanvas
        do begin
             // draw the 4 lines
             Polyline([TL, TR]);
             Polyline([RT, RB]);
             Polyline([BR, BL]);
             Polyline([LB, LT]);
             // now draw the 4 arcs in the corners, being careful to get the order of the endpoints correct!
             Arc(ARect.Left - ARC_RADIUS, ARect.Top - ARC_RADIUS, TL.X, LT.Y, LT.X, LT.Y, TL.X, TL.Y);
             Arc(TR.X, ARect.Top - ARC_RADIUS, ARect.Right + ARC_RADIUS, RT.Y, TR.X, TR.Y, RT.X, RT.Y);
             Arc(BR.X, RB.Y, ARect.Right + ARC_RADIUS, ARect.Bottom + ARC_RADIUS, RB.X, RB.Y, BR.X, BR.Y);
             Arc(ARect.Left - ARC_RADIUS, LB.Y, BL.X, ARect.Bottom + ARC_RADIUS, BL.X, BL.Y, LB.X, LB.Y);
           end; //with ACanvas
       end; // if the rect is big enough for the corner arcs
end; //procedure ClearRectangle

procedure Circle(ACanvas: TCanvas; const Centre: TPoint; const Radius: integer); overload;
// draw a circle given a centre and the radius
begin
  ACanvas.Ellipse(Centre.X - Radius, Centre.Y - Radius, Centre.X + Radius, Centre.Y + Radius);
end; //procedure Circle

procedure Circle(ACanvas: TCanvas; const Centre: TPoint; const ACircumfPoint: TPoint); overload;
// draw a circle given a centre and any point on the circumference
var
  Radius: integer;
begin
  Radius := DistanceBetweenPoints(Centre, ACircumfPoint);
  ACanvas.Ellipse(Centre.X - Radius, Centre.Y - Radius, Centre.X + Radius, Centre.Y + Radius);
end; //procedure Circle

procedure Circle(ACanvas: TCanvas; const A, B, C: TPoint); overload;
// draw a circle given any 3 points on the circumference of a circle
var
  MidABx, MidABy,
  MidBCx, MidBCy,
  CentreX, CentreY,
  SlopeAB, SlopeABperpendicular,
  SlopeBC, SlopeBCperpendicular,
  Radius: double;
  ABvertical, ABhorizontal, BCvertical, BChorizontal: boolean;
begin
  MidABx := (A.X + B.X) / 2;
  MidABy := (A.Y + B.Y) / 2;
  MidBCx := (B.X + C.X) / 2;
  MidBCy := (B.Y + C.Y) / 2;
  ABvertical := (A.X = B.X);
  ABhorizontal := (A.Y = B.Y);
  BCvertical := (B.X = C.X);
  BChorizontal := (B.Y = C.Y);
  CentreX := 0; CentreY := 0; // avoids compiler warnings
  if ABvertical then CentreY := MidABy
  else if ABhorizontal then CentreX := MidABx
  else begin
         SlopeAB := (B.Y - A.Y) / (B.X - A.X);
         SlopeABperpendicular := -1 / SlopeAB;
       end;
  if BCvertical then CentreY := MidBCy
  else if BChorizontal then CentreX := MidBCx
  else begin
         SlopeBC := (C.Y - B.Y) / (C.X - B.X);
         SlopeBCperpendicular := -1 / SlopeBC;
       end;
  // the equation for any straight line is...
  //     Y = mx + c
  // the equation for the line that's perpendicular to AB is...
  //     Y = (SlopeABperpendicular * (X - MidABx))   +  MidABy
  // the equation for the line that's perpendicular to BC is...
  //     Y = (SlopeBCperpendicular * (X - MidBCx))   +  MidBCy
  // but the perpendiculars coincide at the circle's centre, ie the X and Y coincide
  // therefore at the centre...
  //     (SlopeABperpendicular * (X - MidABx)) + MidABy = (SlopeBCperpendicular * (X - MidBCx)) + MidBCy
  if not (ABhorizontal or BChorizontal)
  then // we don't already know CentreX, so we have to calculate it
       CentreX := (MidBCy - MidABy + (SlopeABperpendicular * MidABx) - (SlopeBCperpendicular * MidBCx))
                   /
                  (SlopeABperpendicular - SlopeBCperpendicular);
  // once we know the centre's X, we can use either of the line-equations to find its Y
  if not (ABvertical or BCvertical)
  then // we don't already know CentreY, so we have to calculate it
       begin
         if ABhorizontal
         then CentreY := (SlopeBCperpendicular * (CentreX - MidBCx)) + MidBCy
         else CentreY := (SlopeABperpendicular * (CentreX - MidABx)) + MidABy;
       end;
  // and the radius is a line from the centre to any of A, B or C, since they are on the circumference
  Radius := Sqrt(Sqr(CentreX - A.X) + Sqr(CentreY - A.Y)); // simple Pythagoras
  // and finally we can draw the circle
  ACanvas.Ellipse(Round(CentreX - Radius), Round(CentreY - Radius), Round(CentreX + Radius), Round(CentreY + Radius));
end; //procedure Circle

function DistanceBetweenPoints(const A,B: TPoint): integer; // should really be a real for accuracy, but an integer will do for here
begin
  Result := Round(Sqrt(Sqr(A.X - B.X) + Sqr(A.Y - B.Y))); // simple Pythagoras
end; // function DistanceBetweenPoints

function BearingFromPointToPoint(const A, B: TPoint): integer; // result is 0..359 - should really be a real for accuracy, but an integer will do for here
begin
  if A.X = B.X
  then // it's North/South
       begin
         if A.Y > B.Y
         then Result := 000
         else Result := 180;
       end
  else if A.Y = B.Y
       then // it's East/West
            begin
              if A.X > B.X
              then Result := 270
              else Result := 090;
            end
  else begin
         Result := Round(RadToDeg(ArcTan2(B.X - A.X, A.Y - B.Y))); // will be -180..+180
         if Result < 0
         then Result := Result + 360;
       end;
end; // function BearingFromPointToPoint

end.

